/*
   ibuffer-file.c : Input buffer handling for files

    Copyright (c) 2002 Timo Sirainen

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "lib.h"
#include "alarm-hup.h"
#include "ibuffer-internal.h"
#include "network.h"

#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

#define I_BUFFER_MIN_SIZE 4096

#define BUFFER_IS_BLOCKING(fbuf) \
	((fbuf)->timeout_msecs != 0)

typedef struct {
	_IBuffer ibuf;

	size_t max_buffer_size;
	uoff_t skip_left;

	int timeout_msecs;
	void (*timeout_func)(void *);
	void *timeout_context;

	unsigned int file:1;
	unsigned int autoclose_fd:1;
} FileIBuffer;

static void _close(_IOBuffer *buf)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;
	_IBuffer *_buf = (_IBuffer *) buf;

	if (fbuf->autoclose_fd && _buf->fd != -1) {
		if (close(_buf->fd) < 0)
			i_error("FileIBuffer.close() failed: %m");
		_buf->fd = -1;
	}
}

static void _destroy(_IOBuffer *buf)
{
	_IBuffer *_buf = (_IBuffer *) buf;

	p_free(_buf->iobuf.pool, _buf->w_buffer);
}

static void _set_max_size(_IOBuffer *buf, size_t max_size)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;

	fbuf->max_buffer_size = max_size;
}

static void _set_blocking(_IOBuffer *buf, int timeout_msecs,
			  void (*timeout_func)(void *), void *context)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;

	fbuf->timeout_msecs = timeout_msecs;
	fbuf->timeout_func = timeout_func;
	fbuf->timeout_context = context;

	net_set_nonblock(fbuf->ibuf.fd, timeout_msecs == 0);

	if (timeout_msecs != 0)
		alarm_hup_init();
}

static void i_buffer_grow(_IBuffer *buf, size_t bytes)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;

	buf->buffer_size = buf->pos + bytes;
	buf->buffer_size =
		buf->buffer_size <= I_BUFFER_MIN_SIZE ? I_BUFFER_MIN_SIZE :
		nearest_power(buf->buffer_size);

	if (fbuf->max_buffer_size > 0 &&
	    buf->buffer_size > fbuf->max_buffer_size)
		buf->buffer_size = fbuf->max_buffer_size;

	buf->buffer = buf->w_buffer =
		p_realloc(buf->iobuf.pool, buf->w_buffer, buf->buffer_size);
}

static void i_buffer_compress(_IBuffer *buf)
{
	memmove(buf->w_buffer, buf->w_buffer + buf->skip, buf->pos - buf->skip);
	buf->pos -= buf->skip;

	if (buf->skip > buf->cr_lookup_pos)
		buf->cr_lookup_pos = 0;
	else
		buf->cr_lookup_pos -= buf->skip;

	buf->skip = 0;
}

static ssize_t _read(_IBuffer *buf)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;
	time_t timeout_time;
	uoff_t read_limit;
	size_t size;
	ssize_t ret;

	if (buf->ibuffer.closed)
		return -1;

	if (fbuf->skip_left > 0) {
		i_assert(buf->skip == buf->pos);

		if (fbuf->file) {
			/* we're a file, so we can lseek() */
			i_buffer_seek(&buf->ibuffer, buf->ibuffer.v_offset);
			if (buf->ibuffer.closed)
				return -1;
		}
	}

	buf->ibuffer.buf_errno = 0;

	if (buf->pos == buf->buffer_size) {
		if (buf->skip > 0) {
			/* remove the unused bytes from beginning of buffer */
                        i_buffer_compress(buf);
		} else if (fbuf->max_buffer_size == 0 ||
			   buf->buffer_size < fbuf->max_buffer_size) {
			/* buffer is full - grow it */
			i_buffer_grow(buf, I_BUFFER_MIN_SIZE);
		}

		if (buf->pos == buf->buffer_size)
			return -2; /* buffer full */
	}

	size = buf->buffer_size - buf->pos;
	if (buf->ibuffer.v_limit > 0) {
		i_assert(buf->ibuffer.v_limit >= buf->ibuffer.v_offset);

		read_limit = buf->ibuffer.v_limit -
			buf->ibuffer.v_offset + fbuf->skip_left;
		if (read_limit <= buf->pos - buf->skip) {
			/* virtual limit reached == EOF */
			return -1;
		}

		read_limit -= buf->pos - buf->skip;
		if (size > read_limit)
			size = read_limit;
	}

	timeout_time = GET_TIMEOUT_TIME(fbuf);

	ret = -1;
	do {
		if (ret == 0 && timeout_time > 0 && time(NULL) > timeout_time) {
			/* timeouted */
			if (fbuf->timeout_func != NULL)
				fbuf->timeout_func(fbuf->timeout_context);
			buf->ibuffer.buf_errno = EAGAIN;
			return -1;
		}

		ret = read(buf->fd, buf->w_buffer + buf->pos, size);
		if (ret == 0) {
			/* EOF */
			buf->ibuffer.buf_errno = 0;
			return -1;
		}

		if (ret < 0) {
			if (errno == EINTR || errno == EAGAIN)
				ret = 0;
			else {
				buf->ibuffer.buf_errno = errno;
				return -1;
			}
		}

		if (ret > 0 && fbuf->skip_left > 0) {
			if (fbuf->skip_left >= (size_t)ret) {
				fbuf->skip_left -= ret;
				ret = 0;
			} else {
				ret -= fbuf->skip_left;
				buf->pos += fbuf->skip_left;
				buf->skip += fbuf->skip_left;
				fbuf->skip_left = 0;
			}
		}
	} while (ret == 0 && BUFFER_IS_BLOCKING(fbuf));

	buf->pos += ret;
	return ret;
}

static void _skip(_IBuffer *buf, uoff_t count)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;

	fbuf->skip_left += count - (buf->pos - buf->skip);
	buf->skip = buf->pos = 0;
	buf->ibuffer.v_offset += count;
}

static void _seek(_IBuffer *buf, uoff_t v_offset)
{
	FileIBuffer *fbuf = (FileIBuffer *) buf;
	uoff_t real_offset;
	off_t ret;

	real_offset = buf->ibuffer.start_offset + v_offset;
	if (real_offset > OFF_T_MAX) {
		buf->ibuffer.buf_errno = EOVERFLOW;
		ret = -1;
	} else {
		ret = lseek(buf->fd, (off_t)real_offset, SEEK_SET);
		if (ret < 0)
			buf->ibuffer.buf_errno = errno;
		else if (ret != (off_t)real_offset) {
			buf->ibuffer.buf_errno = EINVAL;
			ret = -1;
		} else {
			buf->skip = buf->pos = 0;
			fbuf->skip_left = 0;
		}
	}

	if (ret < 0)
                i_buffer_close(&buf->ibuffer);
	else {
		buf->ibuffer.buf_errno = 0;
		buf->ibuffer.v_offset = v_offset;
	}
}

IBuffer *i_buffer_create_file(int fd, Pool pool, size_t max_buffer_size,
			      int autoclose_fd)
{
	FileIBuffer *fbuf;
	struct stat st;

	fbuf = p_new(pool, FileIBuffer, 1);
	fbuf->max_buffer_size = max_buffer_size;
	fbuf->autoclose_fd = autoclose_fd;

	fbuf->ibuf.iobuf.close = _close;
	fbuf->ibuf.iobuf.destroy = _destroy;
	fbuf->ibuf.iobuf.set_max_size = _set_max_size;
	fbuf->ibuf.iobuf.set_blocking = _set_blocking;

	fbuf->ibuf.read = _read;
	fbuf->ibuf.skip_count = _skip;
	fbuf->ibuf.seek = _seek;

	/* get size of fd if it's a file */
	if (fstat(fd, &st) < 0)
		st.st_size = 0;
	else if (S_ISREG(st.st_mode))
		fbuf->file = TRUE;

	return _i_buffer_create(&fbuf->ibuf, pool, fd, 0, (uoff_t)st.st_size);
}
